import numpy as np
import matplotlib.pyplot as plt
from matplotlib.widgets import CheckButtons, TextBox
from matplotlib.gridspec import GridSpec

# Define the EMA low-pass filter function
def emalp(signal, N):
    alpha = 2 / (N + 1)
    ema_low_pass = np.zeros_like(signal)
    ema_low_pass[0] = signal[0]
    for i in range(1, len(signal)):
        ema_low_pass[i] = alpha * signal[i] + (1 - alpha) * ema_low_pass[i - 1]
    return ema_low_pass

# Define the EMA high-pass filter function using EMALP
def emahp(signal, N):
    ema_low_pass = emalp(signal, N)
    ema_high_pass = signal - ema_low_pass
    return ema_low_pass, ema_high_pass

# Define the function for corrected EMALP integration
def emalp_integration(signal, N, fs):
    alpha = 2 / (N + 1)
    scaled_signal = signal * 1 / (fs * alpha)
    return emalp(scaled_signal, N)

# Define the integration function to account for the sampling rate
def integrate(signal, fs):
    return np.cumsum(signal) * (1 / fs)

# Compute the derivative of a given motion signal
def compute_derivative(motion, fs):
    derivative = np.zeros_like(motion)
    derivative[1:-1] = (motion[2:] - motion[:-2]) / (2 / fs)
    derivative[0] = (motion[1] - motion[0]) / (1 / fs)
    derivative[-1] = (motion[-1] - motion[-2]) / (1 / fs)
    return derivative

# Recompute Y axis limits
def adjust_axis_limits(ax, lines):
    y_min, y_max = float('inf'), float('-inf')
    for line in lines.values():
        if line.get_visible():
            y_data = line.get_ydata()
            y_min = min(y_min, np.nanmin(y_data)) if np.isfinite(np.nanmin(y_data)) else y_min
            y_max = max(y_max, np.nanmax(y_data)) if np.isfinite(np.nanmax(y_data)) else y_max
    if np.isfinite(y_min) and np.isfinite(y_max):
        ax.set_ylim(y_min, y_max)
    else:
        ax.set_ylim(-1, 1)

# Initialize the lines
def initialize_lines(ax, sample_numbers):
    return {
        'step_signal': ax.plot(sample_numbers, step_signal, label='Step Signal')[0],
        'ema_high_pass': ax.plot(sample_numbers, ema_high_pass, label='EMAHP Filtered Acceleration')[0],
        'velocity_original_integration': ax.plot(sample_numbers, velocity_original_integration, label='Velocity (Original Integration)')[0],
        'velocity_original_emalp1': ax.plot(sample_numbers, velocity_original_emalp1, label='Velocity (Original EMALP1)')[0],
        'velocity_emahp_integration': ax.plot(sample_numbers, velocity_emahp_integration, label='Velocity (EMAHP Integration)')[0],
        'velocity_emahp_emalp1': ax.plot(sample_numbers, velocity_emahp_emalp1, label='Velocity (EMAHP EMALP1)')[0],
        'motion_original': ax.plot(sample_numbers, motion_original, label='Motion (Original)')[0],
        'motion_original_emalp1_integration': ax.plot(sample_numbers, motion_original_emalp1_integration, label='Motion (Original EMALP1 Integration)')[0],
        'motion_original_integration_emalp2': ax.plot(sample_numbers, motion_original_integration_emalp2, label='Motion (Original Integration EMALP2)')[0],
        'motion_original_emalp1_emalp2': ax.plot(sample_numbers, motion_original_emalp1_emalp2, label='Motion (Original EMALP1 EMALP2)')[0],
        'motion_emahp_integration_integration': ax.plot(sample_numbers, motion_emahp_integration_integration, label='Motion (EMAHP Integration Integration)')[0],
        'motion_emahp_emalp1_integration': ax.plot(sample_numbers, motion_emahp_emalp1_integration, label='Motion (EMAHP EMALP1 Integration)')[0],
        'motion_emahp_integration_emalp2': ax.plot(sample_numbers, motion_emahp_integration_emalp2, label='Motion (EMAHP Integration EMALP2)')[0],
        'motion_emahp_emalp1_emalp2': ax.plot(sample_numbers, motion_emahp_emalp1_emalp2, label='Motion (EMAHP EMALP1 EMALP2)')[0],
        'motion_acceleration': ax.plot(sample_numbers, motion_acceleration, label='Motion Acceleration')[0],
    }

# Global variables
maximum_motion = 100  # in mm
maximum_acceleration = 2  # in m/s^2
fs = 500  # Sampling frequency in Hz
emahp_N = 100  # window of acceleration emahp filter
emalp1_N = 200  # window of first integration emalp filter
emalp2_N = 400  # window of second integration emalp filter
n_samples = 0   # make n_samples a global variable
lines = {}      # the lines dictionary
check = None    # make check boxes global

# Recompute signals based on current parameters
def recompute_signals():
    global step_signal, ema_high_pass, velocity_original_integration, velocity_original_emalp1
    global velocity_emahp_integration, velocity_emahp_emalp1, motion_original, motion_original_emalp1_integration
    global motion_original_integration_emalp2, motion_original_emalp1_emalp2, motion_emahp_integration_integration
    global motion_emahp_emalp1_integration, motion_emahp_integration_emalp2, motion_emahp_emalp1_emalp2, motion_acceleration
    global maximum_motion, maximum_acceleration, fs, emahp_N, emalp1_N, emalp2_N, n_samples

    maximum_motion_m = maximum_motion / 1000
    t_unfiltered = np.sqrt(2 * maximum_motion_m / maximum_acceleration)
    duration = 5 * t_unfiltered

    n_samples = int(duration * fs)
    step_signal = np.ones(n_samples) * maximum_acceleration
    step_signal[0] = 0

    ema_low_pass, ema_high_pass = emahp(step_signal, emahp_N)

    velocity_original_integration = integrate(step_signal, fs)
    velocity_original_emalp1 = emalp_integration(step_signal, emalp1_N, fs)
    velocity_emahp_integration = integrate(ema_high_pass, fs)
    velocity_emahp_emalp1 = emalp_integration(ema_high_pass, emalp1_N, fs)

    motion_original = integrate(velocity_original_integration, fs)
    motion_original_emalp1_integration = integrate(velocity_original_emalp1, fs)
    motion_original_integration_emalp2 = emalp_integration(velocity_original_integration, emalp2_N, fs)
    motion_original_emalp1_emalp2 = emalp_integration(velocity_original_emalp1, emalp2_N, fs)

    motion_emahp_integration_integration = integrate(velocity_emahp_integration, fs)
    motion_emahp_emalp1_integration = integrate(velocity_emahp_emalp1, fs)
    motion_emahp_integration_emalp2 = emalp_integration(velocity_emahp_integration, emalp2_N, fs)
    motion_emahp_emalp1_emalp2 = emalp_integration(velocity_emahp_emalp1, emalp2_N, fs)

    motion_acceleration = compute_derivative(compute_derivative(motion_emahp_emalp1_emalp2, fs), fs)

def submit_callback(text):
    try:
        parse_inputs()
        update_signals_and_plot()
        calculate_and_display_effective_motion()
        calculate_and_display_scaling_factors()
    except ValueError:
        pass

def parse_inputs():
    global maximum_acceleration, maximum_motion, emahp_N, emalp1_N, emalp2_N
    maximum_acceleration = float(text_max_acc.text)
    maximum_motion = float(text_max_motion.text)
    emahp_N = int(text_emahp.text)
    emalp1_N = int(text_emalp1.text)
    emalp2_N = int(text_emalp2.text)

def update_signals_and_plot():
    global n_samples, lines
    recompute_signals()
    sample_numbers = np.arange(n_samples)
    ax.cla()
    ax.grid(True)
    lines = initialize_lines(ax, sample_numbers)
    for i, label in enumerate(labels):
        lines[label].set_visible(check.get_status()[i])
    adjust_axis_limits(ax, lines)
    fig.canvas.draw_idle()
    fig.canvas.flush_events()

def calculate_and_display_effective_motion():
    effective_max_motion = np.max(np.abs(motion_emahp_emalp1_emalp2)) * 1000  # Convert to mm
    effective_max_motion_percent = (effective_max_motion / maximum_motion) * 100
    text_effective_max_motion.set_val(f"{effective_max_motion_percent:.2f}%")

# The scaling factors make EMALP behave initially as time integration
def calculate_and_display_scaling_factors():
    alpha_emalp1 = (emalp1_N + 1) / (2 * fs)
    alpha_emalp2 = (emalp2_N + 1) / (2 * fs)
    text_scaling_factor_emalp1.set_val(f"{alpha_emalp1:.4f}")
    text_scaling_factor_emalp2.set_val(f"{alpha_emalp2:.4f}")

# Create figure and grid spec layout
fig = plt.figure(figsize=(12, 5))
height_ratios = [15, 1, 1, 1, 1, 1]
gs = GridSpec(6, 10, figure=fig, height_ratios=height_ratios)

# Create plot area
ax = fig.add_subplot(gs[0, 5:])
ax.grid(True)

# Define update function before creating CheckButtons
def update(label):
    lines[label].set_visible(not lines[label].get_visible())
    adjust_axis_limits(ax, lines)
    ax.grid(True)
    fig.canvas.draw_idle()
    fig.canvas.flush_events()

# Create TextBox widgets for user input and output
axbox1 = fig.add_subplot(gs[1, 2:4])
text_max_acc = TextBox(axbox1, 'Max Acc (m/s^2)', initial=str(maximum_acceleration))
axbox2 = fig.add_subplot(gs[2, 2:4])
text_max_motion = TextBox(axbox2, 'Max Motion (mm)', initial=str(maximum_motion))
axbox3 = fig.add_subplot(gs[2, 6:8])
text_effective_max_motion = TextBox(axbox3, 'Effective Max Motion (%)', initial='0%')
text_effective_max_motion.set_active(False)
axbox4 = fig.add_subplot(gs[3, 2:4])
text_emahp = TextBox(axbox4, 'EMAHP N', initial=str(emahp_N))
axbox5 = fig.add_subplot(gs[4, 2:4])
text_emalp1 = TextBox(axbox5, 'EMALP1 N', initial=str(emalp1_N))
axbox6 = fig.add_subplot(gs[4, 6:8])
text_scaling_factor_emalp1 = TextBox(axbox6, 'Scaling Factor 1', initial='0')
text_scaling_factor_emalp1.set_active(False)
axbox7 = fig.add_subplot(gs[5, 2:4])
text_emalp2 = TextBox(axbox7, 'EMALP2 N', initial=str(emalp2_N))
axbox8 = fig.add_subplot(gs[5, 6:8])
text_scaling_factor_emalp2 = TextBox(axbox8, 'Scaling Factor 2', initial='0')
text_scaling_factor_emalp2.set_active(False)

# Attach the callback function to the TextBox widgets
text_max_acc.on_submit(submit_callback)
text_max_motion.on_submit(submit_callback)
text_emahp.on_submit(submit_callback)
text_emalp1.on_submit(submit_callback)
text_emalp2.on_submit(submit_callback)

# Initial computation of signals and updating the display fields
recompute_signals()
calculate_and_display_effective_motion()
calculate_and_display_scaling_factors()

sample_numbers = np.arange(n_samples)  # Initial sample numbers for first plot
lines = initialize_lines(ax, sample_numbers)

# Hide all lines initially
for line in lines.values():
    line.set_visible(False)

# Create CheckButtons
ax_rax = fig.add_subplot(gs[0, :4])  # CheckButtons in the first rows and first 5 columns
labels = list(lines.keys())         # Initialize labels
visibility = [line.get_visible() for line in lines.values()]

check = CheckButtons(ax_rax, labels, visibility)
check.on_clicked(update)

plt.show()
